from typing import List, Any, Tuple, Optional, Union

import matplotlib as mpl
import numpy as np
from matplotlib import cm
from matplotlib import pyplot as plt
from matplotlib.axes import Axes

from milkviz.utils import norm_arr, set_spines, set_cbar, doc


@doc
def graph(
        nodes: List,
        edges: List[Tuple[Any, Any]],
        nodes_size: Optional[List[float]] = None,
        nodes_color: Union[List[float], List[str], str, None] = "#69B0AC",
        edges_width: Optional[List[float]] = None,
        edges_color: Union[List[float], List[str], str, None] = "#F7C242",
        nodes_size_range: Optional[Tuple[float, float]] = None,
        nodes_color_range: Optional[Tuple[float, float]] = None,
        edges_width_range: Optional[Tuple[float, float]] = None,
        edges_color_range: Optional[Tuple[float, float]] = None,
        node_cmap: Optional[str] = "Purples",
        edge_cmap: Optional[str] = "RdBu",
        node_color_legend_title: str = "node\ncolor",
        edge_color_legend_title: str = "edge\ncolor",
        sizes: Tuple[float, float] = (100, 1500),
        node_shape: str = "o",
        linewidth: Tuple[float, float] = (1, 10),
        connectionstyle: str = 'arc3,rad=0.2',
        layout: str = "kamada_kawai_layout",
        arrowstyle: str = "-",
        ax: Optional[Axes] = None,
) -> Axes:
    """Graph layout

    Args:
        nodes: The graph nodes, a list of nodes
        edges: The graph data, a list of (source, target)
        nodes_size: [size]
        nodes_color: [hue]
        edges_width: The width array that map to edges width
        edges_color: The color array that map to edges color
        nodes_size_range: Use to remap the nodes size, overwrite the min, max of nodes_size array
        nodes_color_range: Use to remap the nodes colors, overwrite the min, max of nodes_color array
        edges_width_range: Use to remap the edges width, overwrite the min, max of nodes_size array
        edges_color_range: Use to remap the edges colors, overwrite the min, max of edges_color array
        node_cmap: The colormap for node
        edge_cmap: The colormap for edge
        node_color_legend_title: Title for nodes colorbar
        edge_color_legend_title: Title for edges colorbar
        sizes: [sizes]
        node_shape: The shape of the node. Specification is as matplotlib.scatter marker, one of ‘so^>v<dph8’.
        linewidth: Line width of symbol border
        connectionstyle: Pass the connectionstyle parameter to create curved arc of rounding
            radius rad. For example, connectionstyle='arc3,rad=0.2'.
            See `matplotlib.patches.ConnectionStyle <https://matplotlib.org/stable/api/_as_gen/matplotlib.patches.ConnectionStyle.html#matplotlib.patches.ConnectionStyle>`_ and
            `matplotlib.patches.FancyArrowPatch <https://matplotlib.org/stable/api/_as_gen/matplotlib.patches.FancyArrowPatch.html#matplotlib.patches.FancyArrowPatch>`_ for more info.
        layout: See
            `networkx.drawing.layout <https://networkx.org/documentation/stable/reference/drawing.html#module-networkx.drawing.layout>`_
        arrowstyle: For directed graphs and arrows==True defaults to ‘-|>’, See
            `matplotlib.patches.ArrowStyle <https://matplotlib.org/stable/api/_as_gen/matplotlib.patches.ArrowStyle.html#matplotlib.patches.ArrowStyle>`_ for more options.
        ax: [ax]

    Returns:
        [return_obj]

    """
    try:
        import networkx as nx
    except ImportError:
        raise ImportError("Extra dependencies needed, Try `pip install networkx`")

    G = nx.Graph()
    G.add_nodes_from(nodes)
    G.add_edges_from(edges)
    if ax is None:
        ax = plt.gca()

    if layout == "bipartite_layout":
        pos = getattr(nx.drawing.layout, layout).__call__(G, [e[0] for e in edges])
    else:
        pos = getattr(nx.drawing.layout, layout).__call__(G)

    if nodes_size is not None:
        vmin, vmax = (None, None) if nodes_size_range is None else nodes_size_range
        nodes_size = norm_arr(nodes_size, size=sizes, vmin=vmin, vmax=vmax)
    else:
        nodes_size = sizes[0]

    if edges_width is not None:
        vmin, vmax = (None, None) if edges_width_range is None else edges_width_range
        edges_width = norm_arr(edges_width, size=linewidth, vmin=vmin, vmax=vmax)
    else:
        edges_width = linewidth[0]

    node_cmin, node_cmax = None, None
    edge_cmin, edge_cmax = None, None
    if nodes_color is not None:
        if not isinstance(nodes_color, str):
            if nodes_color_range is None:
                node_cmin, node_cmax = (np.nanmin(nodes_color), np.nanmax(nodes_color))
            else:
                node_cmin, node_cmax = nodes_color_range

    if edges_color is not None:
        if not isinstance(edges_color, str):
            if edges_color_range is None:
                edge_cmin, edge_cmax = (np.nanmin(edges_color), np.nanmax(edges_color))
            else:
                edge_cmin, edge_cmax = edges_color_range
    nodes_patches = nx.draw_networkx_nodes(G, pos,
                                           node_size=nodes_size,
                                           node_color=nodes_color,
                                           node_shape=node_shape,
                                           vmin=node_cmin,
                                           vmax=node_cmax,
                                           cmap=cm.get_cmap(node_cmap),
                                           linewidths=0,
                                           )
    edges_patches = nx.draw_networkx_edges(G, pos,
                                           width=edges_width,
                                           edge_color=edges_color,
                                           edge_vmin=edge_cmin,
                                           edge_vmax=edge_cmax,
                                           edge_cmap=cm.get_cmap(edge_cmap),
                                           arrowstyle=arrowstyle,
                                           connectionstyle=connectionstyle,
                                           # for safety
                                           node_size=nodes_size,
                                           node_shape=node_shape,
                                           )
    label_options = {"ec": "k", "fc": "white", "alpha": 0.7}
    labels = nx.draw_networkx_labels(G, pos, font_size=mpl.rcParams['font.size'], bbox=label_options)

    bbox_loc = {0: (1.07, 0, 0.1, 0.3), 1: (1.27, 0, 0.1, 0.3)}
    cbar_count = 0
    if node_cmin is not None:
        set_cbar(ax, patches=nodes_patches, bbox=bbox_loc[cbar_count], title=node_color_legend_title,
                 cmin=node_cmin, cmax=node_cmax, cmap=node_cmap)
    cbar_count += 1
    if edge_cmin is not None:
        set_cbar(ax, patches=edges_patches, bbox=bbox_loc[cbar_count], title=edge_color_legend_title,
                 cmin=edge_cmin, cmax=edge_cmax, cmap=edge_cmap)
    set_spines(ax)
    plt.tight_layout()
    return ax



from typing import List, Tuple, Optional, Union, Any, Dict

import matplotlib as mpl
import matplotlib.axes
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.collections import PatchCollection, LineCollection
from matplotlib.colors import Colormap
from natsort import natsorted

from milkviz.utils import set_cbar, set_ticks, set_spines, \
    color_mapper_cat, rotate_points, set_category_legend, color_mapper_val, doc, create_cmap, normalize


@doc
def point_map(
        x: Union[List[float], np.ndarray],
        y: Union[List[float], np.ndarray],
        types: Union[List[str], np.ndarray, None] = None,
        types_colors: Optional[Dict[str, str]] = None,
        values: Union[List[float], np.ndarray, None] = None,
        links: Union[List[Tuple[int, int]], np.ndarray, None] = None,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        colors: Optional[List[Any]] = None,
        cmap: Union[str, Colormap] = None,
        legend_title: Optional[str] = None,
        rotate: Optional[int] = None,
        markersize: Optional[int] = 5,
        linecolor: Union[str, List[str]] = "#cccccc",
        linewidth: int = 1,
        no_spines: bool = True,
        ax: Optional[mpl.axes.Axes] = None,
):
    """Point map

    Args:
        x: The x-coord array
        y: The y-coord array
        types: [types] of points
        types_colors: A dictionary that tells color for every type, key is the type and value is the color
        values: [values] of points
        links: The links between points, should be a list of (point_index_1, point_index_2)
        colors: [hue]
        cmap: [cmap]
        legend_title: [legend_title]
        rotate: The degree to rotate the whole plot according to origin
        markersize: The size of marker
        linecolor: The color of lines
        linewidth: The width of lines
        no_spines: [no_spines]
        ax: [ax]

    Returns:
        [return_obj]

    """
    if ax is None:
        ax = plt.gca()
    if no_spines:
        set_spines(ax)
    set_ticks(ax)
    ax.set_aspect("equal")
    ax.set_xticklabels([])
    ax.set_yticklabels([])
    if rotate is not None:
        x, y = rotate_points(x, y, (0, 0), rotate)

    if links is not None:
        if not isinstance(linecolor, str):
            if len(linecolor) != len(links):
                raise ValueError("Length of linecolor must match to links")
        lines = [[(x[i1], y[i1]), (x[i2], y[i2])] for i1, i2 in links]
        line_collections = LineCollection(lines, linewidths=linewidth, edgecolors=linecolor, zorder=-100)
        ax.add_collection(line_collections)

    if types is not None:
        cmap = "tab20" if cmap is None else cmap
        legend_title = "type" if legend_title is None else legend_title
        cmapper = types_colors
        color_array = None if cmapper is None else [cmapper[t] for t in types]
        if cmapper is None:
            if colors is not None:
                cmapper = color_mapper_cat(types, c_array=colors)
            else:
                cmapper = color_mapper_cat(types, cmap=cmap)
            color_array = [cmapper[t] for t in types]

        ax.scatter(x=x, y=y, c=color_array, s=markersize)
        set_category_legend(ax, cmapper, (1.05, 0, 1, 1), legend_title)
    else:
        if values is not None:
            cmap = "OrRd" if cmap is None else cmap
            legend_title = "value" if legend_title is None else legend_title
            values = normalize(values, vmin, vmax)
            if colors is not None:
                vc_mapper = dict(zip(values, colors))
                cmap = create_cmap([vc_mapper[v] for v in sorted(values)])
            p = ax.scatter(x=x, y=y, c=values, s=markersize, cmap=cmap)
            cmin = np.nanmin(values)
            cmax = np.nanmax(values)
            set_cbar(ax, p, (1.07, 0, 0.1, 0.3), legend_title, cmin, cmax)
        else:
            ax.scatter(x=x, y=y, s=markersize)

    return ax


def point_map3d(
        x: Union[List[float], np.ndarray],
        y: Union[List[float], np.ndarray],
        z: Union[List[float], np.ndarray],
        types: Union[List[str], np.ndarray, None] = None,
        types_colors: Optional[Dict[str, str]] = None,
        values: Union[List[float], np.ndarray, None] = None,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        colors: Optional[List[Any]] = None,
        cmap: Union[str, Colormap] = None,
        legend_title: Optional[str] = None,
        markersize: Optional[int] = 5,
        ax: Optional[mpl.axes.Axes] = None,
):
    """Point map in 3D

        Args:
            x: The x-coord array
            y: The y-coord array
            z: The z-coord array
            types: [types] of points
            types_colors: A dictionary that tells color for every type, key is the type and value is the color
            values: [values] of points
            colors: [hue]
            cmap: [cmap]
            legend_title: [legend_title]
            markersize: The size of marker
            ax: [ax]

        Returns:
            [return_obj]

        """
    if ax is None:
        fig = plt.gcf()
        ax = fig.add_subplot(projection='3d')
    else:
        if not isinstance(ax, Axes3D):
            raise TypeError(f"If you want to use your own axes, initialize it as 3D")
    ax.set_xticklabels([])
    ax.set_yticklabels([])
    ax.set_zticklabels([])
    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")

    if types is not None:
        cmap = "tab20" if cmap is None else cmap
        legend_title = "type" if legend_title is None else legend_title
        cmapper = types_colors
        color_array = None if cmapper is None else [cmapper[t] for t in types]
        if cmapper is None:
            if colors is not None:
                cmapper = color_mapper_cat(types, c_array=colors)
            else:
                cmapper = color_mapper_cat(types, cmap=cmap)
            color_array = [cmapper[t] for t in types]

        ax.scatter(x, y, z, c=color_array, s=markersize)
        set_category_legend(ax, cmapper, (1.2, -0.1, 1, 1), legend_title)
    else:
        if values is not None:
            cmap = "OrRd" if cmap is None else cmap
            legend_title = "value" if legend_title is None else legend_title
            values = normalize(values, vmin, vmax)
            if colors is not None:
                vc_mapper = dict(zip(values, colors))
                cmap = create_cmap([vc_mapper[v] for v in sorted(values)])
            p = ax.scatter(x, y, z, c=values, s=markersize, cmap=cmap)
            cmin = np.nanmin(values)
            cmax = np.nanmax(values)
            set_cbar(ax, p, (1.2, 0.05, 0.1, 0.3), legend_title, cmin, cmax)
        else:
            ax.scatter(x, y, z, s=markersize)

    return ax





@doc
def polygon_map(
        polygons: List[List[Tuple[float, float]]],
        types: Union[List[str], np.ndarray, None] = None,
        values: Union[List[float], np.ndarray, None] = None,
        colors: Optional[List[Any]] = None,
        cmap: Union[str, Colormap] = None,
        legend_title: Optional[str] = None,
        rotate: Optional[int] = None,
        no_spines: bool = True,
        ax: Optional[mpl.axes.Axes] = None,
):
    """Polygon map

    Args:
        polygons: A list of polygons, a polygon is represented by a list of points
        types: [types] of polygons
        values: [values] of polygons
        colors: [hue]
        cmap: [cmap]
        legend_title: [legend_title]
        rotate: The degree to rotate the whole plot according to origin
        no_spines: [no_spines]
        ax: [ax]

    Returns:
        [return_obj]

    """
    polygons = [np.array(polygon) for polygon in polygons]
    vstack_poly = np.vstack(polygons)
    xmin, ymin = np.min(vstack_poly, axis=0)
    xmax, ymax = np.max(vstack_poly, axis=0)

    if rotate is not None:
        rotated_polygons = []
        for polygon in polygons:
            x, y = rotate_points(polygon[:, 0], polygon[:, 1], (0, 0), rotate)
            rotated_polygons.append([(i, j) for i, j in zip(x, y)])
        polygons = rotated_polygons

    if ax is None:
        ax = plt.gca()
    if no_spines:
        set_spines(ax)
    set_ticks(ax)
    ax.set_aspect("equal")
    ax.set_xticklabels([])
    ax.set_yticklabels([])
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)

    if types is not None:
        cmap = "tab20" if cmap is None else cmap
        legend_title = "type" if legend_title is None else legend_title
        uni_types = natsorted(np.unique(types))
        cmapper, color_array = None, None
        if colors is not None:
            cmapper = color_mapper_cat(types, c_array=colors)
            color_array = colors
        else:
            cmapper = color_mapper_cat(types, cmap=cmap)
            color_array = [cmapper[t] for t in types]
        patches = [mpatches.Polygon(polygon) for polygon in polygons]
        patches_collections = PatchCollection(patches, facecolors=[cmapper[t] for t in types])
        ax.add_collection(patches_collections)
        set_category_legend(ax, cmapper, (1.05, 0, 1, 1), legend_title)
    else:
        if values is not None:
            cmap = "OrRd" if cmap is None else cmap
            legend_title = "value" if legend_title is None else legend_title
            if colors is not None:
                vc_mapper = dict(zip(values, colors))
                cmap = create_cmap([vc_mapper[v] for v in sorted(values)])
            colors = color_mapper_val(values, cmap=cmap)
            patches = [mpatches.Polygon(polygon, color=c) for polygon, c in zip(polygons, colors)]
            patches_collections = PatchCollection(patches, facecolors=colors, cmap=cmap)
            ax.add_collection(patches_collections)
            cmin = np.nanmin(values)
            cmax = np.nanmax(values)
            set_cbar(ax, patches_collections, (1.07, 0, 0.1, 0.3), legend_title, cmin, cmax)
        else:
            patches = [mpatches.Polygon(polygon) for polygon in polygons]
            patches_collections = PatchCollection(patches)
            ax.add_collection(patches_collections)

    return ax



from typing import Optional, List, Union, Tuple, Any

import matplotlib as mpl
import matplotlib.axes
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import PatchCollection
from matplotlib.colors import Colormap

from milkviz._dot import set_dot_grid
from milkviz.utils import doc, norm_arr, set_cbar, set_size_legend


@doc
def dot_heatmap(
        dot_size: np.ndarray,
        dot_hue: np.ndarray,
        matrix_hue: Optional[np.ndarray] = None,
        xticklabels: Optional[List[str]] = None,
        yticklabels: Optional[List[str]] = None,
        xlabel: Optional[str] = None,
        ylabel: Optional[str] = None,
        sizes: Tuple[int, int] = (1, 500),
        size_dtype: Any = None,
        dot_cmap: Union[str, Colormap, None] = "RdBu",
        matrix_cmap: Union[str, Colormap, None] = "YlGn",
        size_legend_title: str = "Dot size",
        hue_cbar_title: str = "Dot hue",
        hue_cbar_ticklabels: Optional[List[str]] = None,
        hue_cbar_pos: Optional[Tuple] = None,
        matrix_cbar_title: str = "Matrix",
        matrix_cbar_ticklabels: Optional[List[str]] = None,
        matrix_cbar_pos: Optional[Tuple] = None,
        no_spines: bool = True,
        no_ticks: bool = True,
        ax: Optional[mpl.axes.Axes] = None,
) -> mpl.axes.Axes:
    """Dot heatmap + Matrix heatmap

    Args:
        dot_size: [size]
        dot_hue: [hue]
        matrix_hue: The color array that map to matrix colors
        xticklabels: [xticklabels]
        yticklabels: [yticklabels]
        xlabel: [xlabel]
        ylabel: [ylabel]
        sizes: [sizes]
        size_dtype: [size_dtype]
        dot_cmap: The colormap for dot
        matrix_cmap: The colormap for matrix
        size_legend_title: [size_legend_title]
        hue_cbar_title: [cbar_title] of dot hue
        matrix_cbar_title: [cbar_title] of matrix hue
        hue_cbar_pos: Default is (1.07, 0, 0.1, 0.3)
        matrix_cbar_pos: Default is (1.27, 0, 0.1, 0.3)
        hue_cbar_ticklabels: Text to put on ticks of dot hue colorbar
        matrix_cbar_ticklabels: Text to put on ticks of matrix hue colorbar
        no_spines: [no_spines]
        no_ticks: [no_ticks]
        ax: [ax]

    Returns:
        [return_obj]

    """
    ax, xcoord, ycoord = set_dot_grid(dot_size, ax=ax, xlabel=xlabel, ylabel=ylabel,
                                      xticklabels=xticklabels, yticklabels=yticklabels,
                                      no_spines=no_spines, no_ticks=no_ticks, min_side=3)

    circ_size = norm_arr(dot_size, sizes)
    circ_colors = dot_hue.flatten()
    circ_cmin = np.nanmin(circ_colors)
    circ_cmax = np.nanmax(circ_colors)
    circles = plt.scatter(xcoord, ycoord, s=circ_size, c=circ_colors, cmap=dot_cmap)
    # adding dot size legend
    set_size_legend(ax, dot_size, circ_size, (1.05, 0, 1, 1), size_legend_title, dtype=size_dtype)

    # adding dot colorbar
    hue_cbar_pos = (1.07, 0, 0.1, 0.3) if hue_cbar_pos is None else hue_cbar_pos
    matrix_cbar_pos = (1.27, 0, 0.1, 0.3) if matrix_cbar_pos is None else matrix_cbar_pos
    set_cbar(ax, circles, hue_cbar_pos, hue_cbar_title, circ_cmin, circ_cmax, hue_cbar_ticklabels)

    if matrix_hue is not None:
        rects = [mpatches.Rectangle((rx - 0.5, ry - 0.5), 1, 1) for rx, ry in zip(xcoord, ycoord)]
        rects_colors = matrix_hue.flatten()
        rects_cmin = np.nanmin(rects_colors)
        rects_cmax = np.nanmax(rects_colors)
        rects = PatchCollection(rects, array=rects_colors, cmap=matrix_cmap)
        rects.set_clim(rects_cmin, rects_cmax)
        rects.set_zorder(-1)
        ax.add_collection(rects)

        set_cbar(ax, rects, matrix_cbar_pos, matrix_cbar_title, rects_cmin, rects_cmax, matrix_cbar_ticklabels)
    plt.tight_layout()
    return ax



from typing import Optional, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.axes import Axes
from matplotlib.colors import Colormap

from milkviz.utils import doc
from milkviz.utils import get_cmap_colors, set_category_legend, set_spines


def fold_add(arr):
    new_arr = np.zeros(arr.size)
    new_arr[0] = arr[0]
    for i in range(arr.size - 1):
        new_arr[i + 1] = arr[:i + 2].sum()
    return new_arr.astype(arr.dtype)  # preserve the data type


@doc
def stacked_bar(data: Optional[pd.DataFrame] = None,
                x: Union[str, np.ndarray, None] = None,
                y: Union[str, np.ndarray, None] = None,
                stacked: Union[str, np.ndarray, None] = None,
                orient: str = "v",
                percentage: bool = False,
                cmap: Union[str, Colormap] = "tab20",
                show_values: bool = False,
                ax: Optional[mpl.axes.Axes] = None,
                **kwargs,
                ) -> Axes:
    """Stacked bar plot

    Args:
        data: [data]
        x: [x]
        y: [y]
        stacked: Which columns to plot as stacked type
        orient: "v" or "h"
        percentage: Normalize value to 1
        cmap: [cmap]
        show_values: Whether to display values of each block
        ax: [ax]
        **kwargs: Pass to `seaborn.barplot <https://seaborn.pydata.org/generated/seaborn.barplot.html#seaborn.barplot>`_

    Returns:
        [return_obj]

    """

    if ax is None:
        ax = plt.gca()

    if data is None:
        data = pd.DataFrame({"x": x, "y": y, "stacked": stacked})
        x = "x"
        y = "y"
        stacked = "stacked"

    value_key, label_key = (y, x) if orient == "v" else (x, y)
    stacked_order = data[stacked].unique()
    label_order = data[label_key].unique()

    all_g = []
    for label in label_order:
        g = data[data[label_key] == label].set_index(stacked)
        s_order = []
        for i in stacked_order:
            if i in g.index:
                s_order.append(i)
        g = g.loc[s_order, :]
        v = g[value_key].to_numpy()
        if percentage:
            v = v / v.sum()
        g[value_key] = fold_add(v)
        all_g.append(g)
    # This reverse step is to make the label column in the right order
    data_g = pd.concat(all_g[::-1]).reset_index()

    colors = get_cmap_colors(cmap)
    # This reverse step is to make the stacked column in the right order
    legends_cmapper = {}
    for (n, g), c in zip(data_g.iloc[::-1, :].groupby(stacked, sort=False), colors):
        bar = sns.barplot(x=x, y=y, data=g, ax=ax, color=c, orient=orient, ci=None, **kwargs)
        if show_values:
            for i in range(len(g)):
                text = g.iloc[i, :][value_key]
                if orient == "v":
                    bar.text(i, text, text, ha="center", va="center", bbox=dict(fc="white", alpha=0.7))
                else:
                    bar.text(text, i, text, ha="center", va="center", rotation=-90, bbox=dict(fc="white", alpha=0.7))
        legends_cmapper[n] = c
    set_category_legend(ax, legends_cmapper, (1.05, 0, 1, 1), title=stacked, marker="s")
    set_spines(ax, (1, 0, 1, 0))
    return ax



from typing import Union, List, Any, Optional, Dict, Tuple

import matplotlib as mpl
import numpy as np
from matplotlib import cm
from matplotlib import colors as mcolors
from matplotlib.colors import Colormap
from matplotlib.colors import to_hex
from matplotlib.lines import Line2D
from natsort import natsorted

Colors = Union[str, List[str], Colormap]


def mask_triu(arr, k=1):
    if not isinstance(arr, np.ndarray):
        arr = np.array(arr, dtype=np.float64)
    arr = arr.astype(np.float64)
    shape = arr.shape
    if shape[0] != shape[1]:
        raise ValueError("Must be a square")
    iu = np.triu_indices(shape[0], k=k)
    arr[iu] = np.nan
    return np.fliplr(arr.T)


def norm_arr(arr, size=(1, 100), vmin=None, vmax=None):
    """Normalize the array to a specific range"""
    if not isinstance(arr, np.ndarray):
        arr = np.array(arr)
    arr = arr.flatten()
    add_count = 0
    if vmin is not None:
        arr = np.array([*arr, vmin])
        add_count += 1
    if vmax is not None:
        arr = np.array([*arr, vmax])
        add_count += 1
    amin = np.nanmin(arr)
    amax = np.nanmax(arr)
    size_range = size[1] - size[0]
    arr = (arr - amin) * (size_range / (amax - amin))
    arr += size[0]
    slice_ = -add_count if add_count > 0 else None
    return arr[:slice_]


def adaptive_figsize(size, min_side=5):
    """scale the min number to min_side, while maintain the ratio"""
    size = np.array(size, dtype=np.float64)
    amin = np.amin(size)
    min_ix = np.argmin(size)
    scale = min_side / amin
    if amin < min_side:
        size *= scale
        size[min_ix] = min_side
    return tuple(size)


def get_ax_size(ax):
    x = ax.get_xlim()
    y = ax.get_ylim()
    x_side = x[1] - x[0]
    y_side = y[1] - y[0]
    return x_side, y_side


def get_render_size(ax):
    f = ax.get_figure()
    return f.get_figwidth(), f.get_figheight()


def set_cbar(ax, patches=None, bbox=None, title=None, cmin=None, cmax=None, ticklabels=None, cmap=None):
    axins = ax.inset_axes(bbox, transform=ax.transAxes)
    fig = ax.get_figure()
    if patches is not None:
        cbar = fig.colorbar(patches, cax=axins)
    else:
        cbar = fig.colorbar(cm.ScalarMappable(cmap=cmap), cax=axins)
    cbar.ax.set_title(title, size=mpl.rcParams["font.size"])
    cbar.ax.yaxis.set_tick_params(length=0)  # hide ticks
    if ticklabels is not None:
        cbar.set_ticks(list(np.linspace(cmin, cmax, num=len(ticklabels))), )
        cbar.ax.set_yticklabels(list(ticklabels))


def set_size_legend(ax, size_arr, markersize_arr, bbox, title, dtype=None):
    size_arr = np.asarray(size_arr, dtype=dtype).flatten()
    legend_items = []
    for ratio in [1.0, 0.6, 0.2]:
        if isinstance(size_arr[0], (int, np.integer)) or (dtype == int) or (dtype == np.integer):
            label = int(np.nanmax(size_arr) * ratio)
        else:
            label = round(np.nanmax(size_arr) * ratio, 2)
        item = Line2D((), (), linestyle="none", marker="o", color="black", markerfacecolor="black",
                      label=str(label),
                      markersize=np.sqrt((np.nanmax(markersize_arr))) * ratio
                      )
        legend_items.append(item)
    legend = ax.legend(handles=legend_items, loc="upper left", bbox_to_anchor=bbox,
                       frameon=False, labelspacing=1.2, title=title,
                       fontsize=mpl.rcParams["font.size"])
    legend._legend_box.align = "left"
    ax.add_artist(legend)


def set_category_legend(ax, cmapper, bbox, title, marker="o", reverse=False):
    legend_items = [Line2D((), (), marker=marker, linestyle="none", markersize=8, color=c, markerfacecolor=c, label=t)
                    for t, c in cmapper.items()]
    if reverse:
        legend_items = legend_items[::-1]
    ncol = round(len(legend_items) / 10)
    ncol = ncol if ncol > 0 else 1
    legend = ax.legend(handles=legend_items, loc="upper left", bbox_to_anchor=bbox,
                       frameon=False, title=title, ncol=ncol, columnspacing=0.8,
                       handletextpad=0.1, fontsize=mpl.rcParams["font.size"])
    ax.add_artist(legend)
    return ncol


def set_spines(ax, status=(0, 0, 0, 0)):
    for spine, s in zip(ax.spines.values(), status):
        spine.set_visible(s)


def set_ticks(ax, xticks="none", yticks="none"):
    ax.xaxis.set_ticks_position(xticks)
    ax.yaxis.set_ticks_position(yticks)


def create_cmap(c_array: List[str], name: str = "custom_cmap") -> Colormap:
    return mcolors.ListedColormap(c_array, name=name)


def get_cmap_colors(color: Colors):
    """This utility function allow different color input
    
    1) colormap name
    2) colormap instance
    3) color name, list of color name
    
    """
    cmap = None
    if isinstance(color, str):
        cmap = cm.get_cmap(color)
    elif isinstance(color, Colormap):
        cmap = color
    else:
        return [to_hex(c, keep_alpha=True) for c in color]

    return [to_hex(cmap(i), keep_alpha=True) for i in range(cmap.N)]


def color_mapper_cat(types: Union[List[Any], np.ndarray],
                     c_array: Optional[List[str]] = None,
                     cmap: Optional[Colors] = None,
                     ) -> Dict:
    """
    
    Args:
        types: All input types
        c_array: The color array that maps to the types
        cmap: An array of colors or colormap


    Returns:
        {type: color}
        
    """
    N = len(types)
    uni_types = natsorted(np.unique(types))
    if c_array is not None:
        if len(c_array) < N:
            raise ValueError(f"The length of input color array {len(c_array)} does not match types {N}")
        else:
            cmapper = dict(zip(types, c_array))
            return {k: cmapper[k] for k in uni_types}
    if cmap is not None:
        all_colors = []
        colors_pool = get_cmap_colors(cmap)
        while len(all_colors) < N:
            all_colors += colors_pool

        return dict(zip(uni_types, all_colors))


def color_mapper_val(values: Union[List[Any], np.ndarray],
                     c_array: Optional[List[str]] = None,
                     cmap: Optional[Colors] = None, ) -> List[Tuple]:
    if not isinstance(values, np.ndarray):
        values = np.ndarray(values)
    vmin = np.nanmin(values)
    vmax = np.nanmax(values)
    norm = mcolors.Normalize(vmin=vmin, vmax=vmax)
    if c_array is not None:
        cmap = create_cmap(c_array)
    mapper = cm.ScalarMappable(norm=norm, cmap=cmap)
    return [to_hex(mapper.to_rgba(v), keep_alpha=True) for v in values]



from typing import Optional, List, Union, Tuple

import matplotlib as mpl
import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.colors import Colormap

from milkviz.utils import norm_arr, doc
from milkviz.utils.fig import set_size_legend, set_cbar


@doc
def bubble(data: Optional[pd.DataFrame] = None,
           x: Union[str, np.ndarray, None] = None,
           y: Union[str, np.ndarray, None] = None,
           hue: Union[str, np.ndarray, None] = None,
           size: Union[str, np.ndarray, None] = None,
           cmap: Union[str, Colormap] = "RdBu",
           sizes: Tuple[int, int] = (1, 500),
           size_legend_title: str = "size",
           hue_cbar_title: str = "hue",
           hue_cbar_ticklabels: Optional[List[str]] = None,
           ax: Optional[mpl.axes.Axes] = None,
           ) -> mpl.axes.Axes:
    """Bubble plot

    Args:
        data: [data]
        x: [x]
        y: [y]
        hue: [hue]
        size: [size]
        cmap: [cmap]
        sizes: [sizes]
        size_legend_title: [size_legend_title]
        hue_cbar_title: [cbar_title]
        hue_cbar_ticklabels: Overwrite the ticklabels on colorbar
        ax: [ax]

    Returns:
        [return_obj]

    """
    if data is not None:
        x = data[x].to_numpy()
        y = data[y].to_numpy()
        hue = data[hue].to_numpy() if hue is not None else None
        size = data[size].to_numpy() if size is not None else None

    if size is None:
        raise ValueError("size must be provided")

    if ax is None:
        ax = plt.gca()

    circ_size = norm_arr(size, sizes)
    bubbles = plt.scatter(x, y, s=circ_size, c=hue, cmap=cmap)
    set_size_legend(ax, size, circ_size, bbox=(1.05, 0, 1, 1), title=size_legend_title)
    if hue is not None:
        cmin = np.nanmin(hue)
        cmax = np.nanmax(hue)
        set_cbar(ax, bubbles, (1.07, 0, 0.1, 0.3), hue_cbar_title, cmin, cmax, hue_cbar_ticklabels)

    return ax



from typing import List, Union, Optional, Tuple

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import cm
from matplotlib.colors import Colormap

from milkviz.utils import color_mapper_cat, set_cbar, set_category_legend, doc


def anno_colors(info: pd.DataFrame, color: Union[str, List[str]]):
    uni_types = np.unique(info.to_numpy().flatten())
    colors_mapper = color_mapper_cat(color, uni_types)
    return info.replace(colors_mapper)


def move_bbox(bbox, move_index, amount):
    bbox = list(bbox)
    bbox[move_index] += amount
    return tuple(bbox)


def color_legend_split(ax, legend_pos, legend_contents, colors_mapper, padding):
    extend_ncol = {1: 1, 2: 1.5, 3: 2}
    ncol = 1
    legend_pos = move_bbox(legend_pos, 0, -padding)
    for k, v in legend_contents.items():
        legend_pos = move_bbox(legend_pos, 0, padding)
        cmapper = {i: colors_mapper[i] for i in v}
        padding /= ncol  # reset the padding
        ncol = set_category_legend(ax, cmapper, bbox=legend_pos,
                                   title=k, marker="s")
        padding *= extend_ncol[ncol]  # if ncol > 1, extend the padding


def color_legend(ax, legend_pos, legend_contents, colors_mapper, title=None):
    cmapper = {}
    for k, v in legend_contents.items():
        for i in v:
            cmapper[i] = colors_mapper[i]
    set_category_legend(ax, cmapper, bbox=legend_pos, title=title, marker="s")


@doc
def anno_clustermap(
        data: pd.DataFrame,
        row_colors: Union[str, List[str], None] = None,
        row_label: Optional[str] = None,
        col_colors: Union[str, List[str], None] = None,
        col_label: Optional[str] = None,
        heat_cmap: Union[str, Colormap] = "RdBu_r",
        row_colors_cmap: Union[str, List[str]] = "tab20",
        col_colors_cmap: Union[str, List[str]] = "tab20",
        categorical_cbar: Optional[List[str]] = None,
        cbar_title: str = "data",
        cbar_pos: Optional[Tuple[float, float, float, float]] = None,
        row_legend_pos: Optional[Tuple[float, float, float, float]] = None,
        col_legend_pos: Optional[Tuple[float, float, float, float]] = None,
        row_legend_title: str = "row",
        col_legend_title: str = "col",
        row_legend_split: bool = True,
        col_legend_split: bool = True,
        row_legend_padding: float = 0.25,
        col_legend_padding: float = 0.25,
        **kwargs,
) -> sns.matrix.ClusterGrid:
    """Color or label annotated clustermap

    Args:
        data: [data], multi-levels annotations should store in MultiIndex
        row_colors: The index level used to label rows in color stripe
        row_label: The index level used for x-axis label
        col_colors: The columns level used to label columns in color stripe
        col_label: The columns level used for y-axis label
        heat_cmap: The colormap for heatmap
        row_colors_cmap: The colormap for row_colors
        col_colors_cmap: The colormap for col_colors
        categorical_cbar: Turn the colorbar in to categorical legend in text
        cbar_title: The colorbar title
        cbar_pos: The colorbar position, default is (1.05, 0.0, 0.05, 0.15)
        row_legend_title: The title of row legend, when row_legend_split = False
        col_legend_title: The title of col legend, when col_legend_split = False
        row_legend_pos: row_colors' legend's position, default is (0.02, -0.36, 0.1, 0.3)
        col_legend_pos: col_colors' legend's position, default is (1.05, 0.7, 0.1, 0.3)
        row_legend_split: Whether to split each level of row legend
        col_legend_split: Whether to split each level of col legend
        row_legend_padding: Use to specific the padding between multiple row/col legends
        col_legend_padding: Use to specific the padding between multiple row/col legends
        **kwargs: Pass to `seaborn.clustermap <https://seaborn.pydata.org/generated/seaborn.clustermap.html#seaborn.clustermap>`_

    Returns:
        A `seaborn.matrix.ClusterGrid` instance

    """
    # split the dataframe
    raw_data = data.to_numpy()
    row_info = data.index.to_frame(index=False)
    col_info = data.columns.to_frame(index=False)
    row_label = row_info[row_label] if row_label is not None else None
    col_label = col_info[col_label] if col_label is not None else None
    plot_data = pd.DataFrame(raw_data, index=row_label, columns=col_label)

    clustermap_kwargs = dict(cmap=heat_cmap, cbar_pos=None, **kwargs)
    row_colors_mapper = None
    col_colors_mapper = None
    row_legend_contents = None
    col_legend_contents = None

    if row_colors is not None:
        info = row_info[row_colors]
        row_legend_contents = {rc: np.unique(info[rc].to_numpy().flatten()) for rc in row_colors}
        row_colors_mapper = color_mapper_cat(info.to_numpy().flatten(), cmap=row_colors_cmap)
        clustermap_kwargs["row_colors"] = info.replace(row_colors_mapper)

    if col_colors is not None:
        info = col_info[col_colors]
        col_legend_contents = {cc: np.unique(info[cc].to_numpy().flatten()) for cc in col_colors}
        col_colors_mapper = color_mapper_cat(info.to_numpy().flatten(), cmap=col_colors_cmap)
        clustermap_kwargs["col_colors"] = info.replace(col_colors_mapper)

    g = sns.clustermap(plot_data, **clustermap_kwargs)
    cmin = np.nanmin(plot_data)
    cmax = np.nanmax(plot_data)
    cbar_pos = (1.05, 0.0, 0.05, 0.15) if cbar_pos is None else cbar_pos
    if categorical_cbar is not None:
        cmap = cm.get_cmap(heat_cmap)
        cmap_size = np.linspace(0, cmap.N, len(categorical_cbar), dtype=int)
        cmapper = dict(zip(categorical_cbar, [cmap(i) for i in cmap_size]))
        set_category_legend(g.ax_heatmap, cmapper, bbox=cbar_pos,
                            title=cbar_title, marker="s", reverse=True)
    else:
        set_cbar(g.ax_heatmap, bbox=cbar_pos, cmin=cmin, cmax=cmax, cmap=heat_cmap, title=cbar_title)

    # plot row colors legend
    if row_colors is not None:
        row_legend_pos = (0.02, -0.36, 0.1, 0.3) if row_legend_pos is None else row_legend_pos
        if row_legend_split:
            color_legend_split(g.ax_heatmap, row_legend_pos, row_legend_contents,
                               row_colors_mapper, row_legend_padding)
        else:
            color_legend(g.ax_heatmap, row_legend_pos, row_legend_contents,
                         row_colors_mapper, row_legend_title)
    # plot col colors legend
    if col_colors is not None:
        col_legend_pos = (1.05, 0.7, 0.1, 0.3) if col_legend_pos is None else col_legend_pos
        if col_legend_split:
            color_legend_split(g.ax_heatmap, col_legend_pos, col_legend_contents,
                               col_colors_mapper, col_legend_padding)
        else:
            color_legend(g.ax_heatmap, col_legend_pos, col_legend_contents,
                         col_colors_mapper, col_legend_title)

    # close labels and ticks on y-axis
    if row_label is None:
        g.ax_heatmap.set_yticklabels("")
        g.ax_heatmap.set_ylabel("")
        g.ax_heatmap.set_yticks([])

    # close labels and ticks on x-axis
    if col_label is None:
        g.ax_heatmap.set_xticklabels("")
        g.ax_heatmap.set_xlabel("")
        g.ax_heatmap.set_xticks([])

    return g



from typing import Optional, List, Union, Tuple, Any

import matplotlib as mpl
import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np

from milkviz.utils import adaptive_figsize, norm_arr, doc, set_size_legend, set_spines, set_ticks


def set_dot_grid(data,
                 ax=None,
                 xlabel=None,
                 ylabel=None,
                 xticklabels=None,
                 yticklabels=None,
                 no_spines=True,
                 no_ticks=True,
                 min_side=4,
                 ):
    Y, X = data.shape
    x, y = np.meshgrid(np.arange(X), np.arange(Y))  # Get the coordinates
    xcoord = x.flatten()
    ycoord = y.flatten()

    if ax is None:
        ratio = 0.6
        figsize = adaptive_figsize((X * ratio, Y * ratio), min_side=min_side)
        _, ax = plt.subplots(figsize=figsize)
    if no_spines:
        set_spines(ax)
    if no_ticks:
        set_ticks(ax)
    ax.set_aspect("equal")
    ax.set_xlim(-0.5, X - 0.5)
    ax.set_ylim(-0.5, Y - 0.5)
    ax.set(
        xlabel=xlabel,
        ylabel=ylabel,
        xticks=np.arange(X),
        yticks=np.arange(Y),
        xticklabels=xticklabels if xticklabels is not None else [],
        yticklabels=yticklabels if yticklabels is not None else [],
    )
    plt.xticks(rotation=90)

    return ax, xcoord, ycoord


@doc
def dot(
        dot_size: np.ndarray,
        dot_hue: Union[str, np.ndarray, None] = None,
        xticklabels: Optional[List[str]] = None,
        yticklabels: Optional[List[str]] = None,
        xlabel: Optional[str] = None,
        ylabel: Optional[str] = None,
        sizes: Tuple[int, int] = (1, 500),
        size_dtype: Any = None,
        legend_title: str = "size",
        no_spines: bool = True,
        no_ticks: bool = True,
        ax: Optional[mpl.axes.Axes] = None,
) -> mpl.axes.Axes:
    """Dot plot

    Args:
        dot_size: [size]
        dot_hue: [hue]
        xticklabels: [xticklabels]
        yticklabels: [yticklabels]
        xlabel: [xlabel]
        ylabel: [ylabel]
        sizes: [sizes]
        size_dtype: [size_dtype]
        legend_title: [size_legend_title]
        no_spines: [no_spines]
        no_ticks: [no_ticks]
        ax: [ax]

    Returns:
        [return_obj]

    """

    ax, xcoord, ycoord = set_dot_grid(dot_size, ax=ax, xlabel=xlabel, ylabel=ylabel,
                                      xticklabels=xticklabels, yticklabels=yticklabels,
                                      no_spines=no_spines, no_ticks=no_ticks, min_side=3)

    if dot_hue is not None:
        if isinstance(dot_hue, str):
            circ_colors = np.repeat(dot_hue, dot_size.size)
        else:
            if np.array((dot_hue.shape == dot_size.shape)).all():
                circ_colors = dot_hue.flatten()
            else:
                raise ValueError("dot_hue does not match the shape of dot_size")
    else:
        circ_colors = np.repeat("#D75455", dot_size.size)

    circ_size = norm_arr(dot_size, sizes)
    _ = plt.scatter(xcoord, ycoord, s=circ_size, c=circ_colors)
    # adding dot size legend
    set_size_legend(ax, dot_size, circ_size, (1.05, 0, 1, 1), legend_title, dtype=size_dtype)
    ax.grid(False)
    return ax



from .doc import doc
from .fig import adaptive_figsize, norm_arr, set_cbar, set_size_legend, set_category_legend, \
    set_ticks, set_spines, color_mapper_cat, color_mapper_val, mask_triu, get_cmap_colors, create_cmap, get_render_size
from .geo import rotate_points, normalize



from typing import Optional, List, Any

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.figure import Figure

from milkviz.utils import doc


@doc
def upset(
        data: List[List[Any]],
        names: Optional[List[str]] = None,
        orient: str = "v",
        min_subset_size: Optional[int] = None,
        max_subset_size: Optional[int] = None,
        min_degree: Optional[int] = None,
        max_degree: Optional[int] = None,
        show_counts: bool = True,
        show_percentages: bool = False,
        fig: Optional[Figure] = None,
        **kwargs,
) -> Figure:
    """Upset plot for intersections between arbitary sets

    This is simply a wrapper around `upsetplot <https://github.com/jnothman/UpSetPlot>`_

    Unlike `venn`, this will not consider duplicates

    Args:
        data: A series of anything, set will be auto computed for it
        names: The name for each set
        orient: "v" or "h"
        {min, max}_subset_size: Minimum/Maximum size threshold of a subset to be shown in the plot.
        {min, max}_degree: Minimum/Maximum degree of a subset to be shown in the plot
        show_counts: Whether to label the intersection size bars with the cardinality of the intersection.
        show_percentages: Whether to label the intersection size bars with the percentage of the intersection relative to the total dataset.

        **kwargs: Pass to `upsetplot.UpSet <https://upsetplot.readthedocs.io/en/stable/api.html#upsetplot.UpSet>`_

    Returns:
        A `matplotlib.figure.Figure` object

    """
    try:
        from upsetplot import UpSet
        from upsetplot import from_contents
    except ImportError:
        raise ImportError("Extra dependencies needed, Try `pip install UpSetPlot`")

    if fig is None:
        fig = plt.gcf()
    params = dict(
        orientation="vertical" if orient == "h" else "horizontal",
        min_subset_size=min_subset_size,
        max_subset_size=max_subset_size,
        min_degree=min_degree,
        max_degree=max_degree,
        show_counts=show_counts,
        show_percentages=show_percentages,
        **kwargs
    )
    if names is None:
        names = [f"Set{i}" for i in range(len(data))]
    data = [np.unique(d) for d in data]
    contents = dict(zip(names, data))
    data = from_contents(contents)
    p = UpSet(data, **params).plot(fig=fig)
    # No a good thing to do here to prevent jupyter to show the plot twice

    return fig



from ._bubble import bubble
from ._cell_map import point_map, polygon_map, point_map3d
from ._clustermap import anno_clustermap
from ._dot import dot
from ._dot_matrix import dot_heatmap
from ._graph import graph
from ._stacked_bar import stacked_bar
from ._upset import upset
from ._venn import venn
from .utils import mask_triu
